#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2014 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Fee implementation
"""

import time
import random

from persistent import Persistent

from zope.annotation.interfaces import IAnnotations
from zope.component import adapter
from zope.component import adapts
from zope.container.btree import BTreeContainer
from zope.container.contained import Contained
from zope.container.contained import NameChooser
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.interface import implements
from zope.lifecycleevent.interfaces import IObjectAddedEvent

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.group.interfaces import IGroup
from schooltool.person.interfaces import IPerson
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.schoolyear.subscriber import ObjectEventAdapterSubscriber
from schooltool.stream.interfaces import IStream

from schooltool.fee.interfaces import IReceiptNumbers
from schooltool.fee.interfaces import IAmount
from schooltool.fee.interfaces import ICredit
from schooltool.fee.interfaces import IDebit
from schooltool.fee.interfaces import IFeeContainer
from schooltool.fee.interfaces import IIncomeGeneratingActivityContainer
from schooltool.fee.interfaces import ISchoolYearFeeContainer
from schooltool.fee.interfaces import IGroupFeeContainer
from schooltool.fee.interfaces import IStreamFeeContainer


PERSON_FEE_KEY = 'schooltool.fee.person'
SCHOOLYEAR_FEE_KEY = 'schooltool.fee.schoolyear'
GROUP_FEE_KEY = 'schooltool.fee.group'
STREAM_FEE_KEY = 'schooltool.fee.stream'
RECEIPT_NUMBERS_KEY = 'schooltool.fee.receipt_numbers'


class FeeContainer(BTreeContainer):

    implements(IFeeContainer)


class Amount(Persistent, Contained):

    implements(IAmount)

    title = None
    description = None
    date = None
    amount = None


class Credit(Amount):

    implements(ICredit)

    receipt = None

    @property
    def balance_amount(self):
        return self.amount


class Debit(Amount):

    implements(IDebit)

    last_date = None
    label = None
    master = None

    @property
    def balance_amount(self):
        return -self.amount


@implementer(IFeeContainer)
@adapter(IPerson)
def getPersonFeeContainer(person):
    ann = IAnnotations(person)
    try:
        return ann[PERSON_FEE_KEY]
    except (KeyError,):
        ann[PERSON_FEE_KEY] = FeeContainer()
        ann[PERSON_FEE_KEY].__name__ = 'fees'
        ann[PERSON_FEE_KEY].__parent__ = person
        return ann[PERSON_FEE_KEY]


@implementer(IPerson)
@adapter(IFeeContainer)
def getFeeContainerPerson(container):
    if IPerson.providedBy(container.__parent__):
        return container.__parent__
    if IIncomeGeneratingActivityContainer.providedBy(container.__parent__):
        return container


@implementer(IPerson)
@adapter(IAmount)
def getAmountPerson(amount):
    container = amount.__parent__
    return IPerson(container)


@implementer(IFeeContainer)
@adapter(ISchoolYear)
def getSchoolYearFeeContainer(schoolyear):
    ann = IAnnotations(schoolyear)
    try:
        return ann[SCHOOLYEAR_FEE_KEY]
    except (KeyError,):
        ann[SCHOOLYEAR_FEE_KEY] = container = FeeContainer()
        alsoProvides(container, ISchoolYearFeeContainer)
        ann[SCHOOLYEAR_FEE_KEY].__name__ = 'fees'
        ann[SCHOOLYEAR_FEE_KEY].__parent__ = schoolyear
        return ann[SCHOOLYEAR_FEE_KEY]


@implementer(IFeeContainer)
@adapter(IGroup)
def getGroupFeeContainer(group):
    ann = IAnnotations(group)
    try:
        return ann[GROUP_FEE_KEY]
    except (KeyError,):
        ann[GROUP_FEE_KEY] = container = FeeContainer()
        alsoProvides(container, IGroupFeeContainer)
        ann[GROUP_FEE_KEY].__name__ = 'fees'
        ann[GROUP_FEE_KEY].__parent__ = group
        return ann[GROUP_FEE_KEY]


@implementer(IFeeContainer)
@adapter(IStream)
def getStreamFeeContainer(stream):
    ann = IAnnotations(stream)
    try:
        return ann[GROUP_FEE_KEY]
    except (KeyError,):
        ann[GROUP_FEE_KEY] = container = FeeContainer()
        alsoProvides(container, IStreamFeeContainer)
        ann[GROUP_FEE_KEY].__name__ = 'fees'
        ann[GROUP_FEE_KEY].__parent__ = stream
        return ann[GROUP_FEE_KEY]


class SchoolYearFeeContainerNameChooser(NameChooser):

    adapts(ISchoolYearFeeContainer)

    def chooseName(self, name, obj):
        schoolyear = self.context.__parent__
        n = u'SchoolYearFee.%s.%s.%s' % (
            schoolyear.__name__, time.time(), random.random())
        self.checkName(n, obj)
        return n


class ReceiptNumbers(Persistent):

    last = 1

    def __unicode__(self):
        return '%06d' % self.last


@implementer(IReceiptNumbers)
@adapter(ISchoolToolApplication)
def getApplicationReceiptNumbers(app):
    ann = IAnnotations(app)
    try:
        return ann[RECEIPT_NUMBERS_KEY]
    except (KeyError,):
        ann[RECEIPT_NUMBERS_KEY] = ReceiptNumbers()
        return ann[RECEIPT_NUMBERS_KEY]


class IncrementReceiptNumbers(ObjectEventAdapterSubscriber):

    adapts(IObjectAddedEvent, ICredit)

    def __call__(self):
        app = ISchoolToolApplication(None)
        numbers = IReceiptNumbers(app)
        numbers.last += 1
